import { BatchWriteItemCommand, QueryCommand, } from "@aws-sdk/client-dynamodb";
import { DeleteObjectsCommand, GetObjectCommand, ListObjectsV2Command, PutObjectCommand, } from "@aws-sdk/client-s3";
// //@ts-ignore
// import { getDerivedTags } from "next/dist/server/lib/incremental-cache/utils";
import path from "path";
import { debug, error } from "./logger.js";
// TODO: Remove this, temporary only to run some tests
const getDerivedTags = (tags) => tags;
// Expected environment variables
const { CACHE_BUCKET_NAME, CACHE_BUCKET_KEY_PREFIX, CACHE_DYNAMO_TABLE, NEXT_BUILD_ID, } = process.env;
export default class S3Cache {
    client;
    dynamoClient;
    buildId;
    constructor(_ctx) {
        this.client = globalThis.S3Client;
        this.dynamoClient = globalThis.dynamoClient;
        this.buildId = NEXT_BUILD_ID;
    }
    async get(key, options) {
        if (globalThis.disableIncrementalCache) {
            return null;
        }
        const isFetchCache = typeof options === "object" ? options.fetchCache : options;
        const keys = await this.listS3Object(key);
        if (keys.length === 0)
            return null;
        debug("keys", keys);
        return isFetchCache
            ? this.getFetchCache(key, keys)
            : this.getIncrementalCache(key, keys);
    }
    async getFetchCache(key, keys) {
        debug("get fetch cache", { key });
        try {
            const { Body, LastModified } = await this.getS3Object(key, "fetch", keys);
            const lastModified = await this.getHasRevalidatedTags(key, LastModified?.getTime());
            if (lastModified === -1) {
                // If some tags are stale we need to force revalidation
                return null;
            }
            if (Body === null)
                return null;
            return {
                lastModified,
                value: JSON.parse((await Body?.transformToString()) ?? "{}"),
            };
        }
        catch (e) {
            error("Failed to get fetch cache", e);
            return null;
        }
    }
    async getIncrementalCache(key, keys) {
        if (keys.includes(this.buildS3Key(key, "body"))) {
            debug("get body cache ", { key });
            try {
                const [{ Body, LastModified }, { Body: MetaBody }] = await Promise.all([
                    this.getS3Object(key, "body", keys),
                    this.getS3Object(key, "meta", keys),
                ]);
                const body = await Body?.transformToByteArray();
                const meta = JSON.parse((await MetaBody?.transformToString()) ?? "{}");
                return {
                    lastModified: LastModified?.getTime(),
                    value: {
                        kind: "ROUTE",
                        body: Buffer.from(body ?? Buffer.alloc(0)),
                        status: meta.status,
                        headers: meta.headers,
                    },
                };
            }
            catch (e) {
                error("Failed to get body cache", e);
            }
            return null;
        }
        if (keys.includes(this.buildS3Key(key, "html"))) {
            const isJson = keys.includes(this.buildS3Key(key, "json"));
            const isRsc = keys.includes(this.buildS3Key(key, "rsc"));
            debug("get html cache ", { key, isJson, isRsc });
            if (!isJson && !isRsc)
                return null;
            try {
                const [{ Body, LastModified }, { Body: PageBody }, { Body: MetaBody }] = await Promise.all([
                    this.getS3Object(key, "html", keys),
                    this.getS3Object(key, isJson ? "json" : "rsc", keys),
                    this.getS3Object(key, "meta", keys),
                ]);
                const lastModified = await this.getHasRevalidatedTags(key, LastModified?.getTime());
                if (lastModified === -1) {
                    // If some tags are stale we need to force revalidation
                    return null;
                }
                const meta = JSON.parse((await MetaBody?.transformToString()) ?? "{}");
                return {
                    lastModified,
                    value: {
                        kind: "PAGE",
                        html: (await Body?.transformToString()) ?? "",
                        pageData: isJson
                            ? JSON.parse((await PageBody?.transformToString()) ?? "{}")
                            : await PageBody?.transformToString(),
                        status: meta.status,
                        headers: meta.headers,
                    },
                };
            }
            catch (e) {
                error("Failed to get html cache", e);
            }
            return null;
        }
        // Check for redirect last. This way if a page has been regenerated
        // after having been redirected, we'll get the page data
        if (keys.includes(this.buildS3Key(key, "redirect"))) {
            debug("get redirect cache", { key });
            try {
                const { Body, LastModified } = await this.getS3Object(key, "redirect", keys);
                return {
                    lastModified: LastModified?.getTime(),
                    value: JSON.parse((await Body?.transformToString()) ?? "{}"),
                };
            }
            catch (e) {
                error("Failed to get redirect cache", e);
            }
            return null;
        }
        return null;
    }
    async set(key, data) {
        if (globalThis.disableIncrementalCache) {
            return;
        }
        if (data?.kind === "ROUTE") {
            const { body, status, headers } = data;
            await Promise.all([
                this.putS3Object(key, "body", body),
                this.putS3Object(key, "meta", JSON.stringify({ status, headers })),
            ]);
        }
        else if (data?.kind === "PAGE") {
            const { html, pageData } = data;
            const isAppPath = typeof pageData === "string";
            let metaPromise = Promise.resolve();
            if (data.status || data.headers) {
                metaPromise = this.putS3Object(key, "meta", JSON.stringify({ status: data.status, headers: data.headers }));
            }
            await Promise.all([
                this.putS3Object(key, "html", html),
                this.putS3Object(key, isAppPath ? "rsc" : "json", isAppPath ? pageData : JSON.stringify(pageData)),
                metaPromise,
            ]);
        }
        else if (data?.kind === "FETCH") {
            await this.putS3Object(key, "fetch", JSON.stringify(data));
        }
        else if (data?.kind === "REDIRECT") {
            // delete potential page data if we're redirecting
            await this.deleteS3Objects(key);
            await this.putS3Object(key, "redirect", JSON.stringify(data));
        }
        else if (data === null || data === undefined) {
            await this.deleteS3Objects(key);
        }
        // Write getDerivedTags to dynamodb
        // If we use an in house version of getDerivedTags in build we should use it here instead of next's one
        const derivedTags = data?.kind === "FETCH"
            ? getDerivedTags(data.data.tags ?? [])
            : data?.kind === "PAGE"
                ? getDerivedTags(data.headers?.["x-next-cache-tags"]?.split(",") ?? [])
                : [];
        debug("derivedTags", derivedTags);
        // Get all tags stored in dynamodb for the given key
        // If any of the derived tags are not stored in dynamodb for the given key, write them
        const storedTags = await this.getTagsByPath(key);
        const tagsToWrite = derivedTags.filter((tag) => !storedTags.includes(tag));
        if (tagsToWrite.length > 0) {
            await this.batchWriteDynamoItem(tagsToWrite.map((tag) => ({
                path: key,
                tag: tag,
            })));
        }
    }
    async revalidateTag(tag) {
        if (globalThis.disableDynamoDBCache || globalThis.disableIncrementalCache) {
            return;
        }
        debug("revalidateTag", tag);
        // Find all keys with the given tag
        const paths = await this.getByTag(tag);
        debug("Items", paths);
        // Update all keys with the given tag with revalidatedAt set to now
        await this.batchWriteDynamoItem(paths?.map((path) => ({
            path: path,
            tag: tag,
        })) ?? []);
    }
    // DynamoDB handling
    async getTagsByPath(path) {
        try {
            if (disableDynamoDBCache)
                return [];
            const result = await this.dynamoClient.send(new QueryCommand({
                TableName: CACHE_DYNAMO_TABLE,
                IndexName: "revalidate",
                KeyConditionExpression: "#key = :key",
                ExpressionAttributeNames: {
                    "#key": "path",
                },
                ExpressionAttributeValues: {
                    ":key": { S: this.buildDynamoKey(path) },
                },
            }));
            const tags = result.Items?.map((item) => item.tag.S ?? "") ?? [];
            debug("tags for path", path, tags);
            return tags;
        }
        catch (e) {
            error("Failed to get tags by path", e);
            return [];
        }
    }
    //TODO: Figure out a better name for this function since it returns the lastModified
    async getHasRevalidatedTags(key, lastModified) {
        try {
            if (disableDynamoDBCache)
                return lastModified ?? Date.now();
            const result = await this.dynamoClient.send(new QueryCommand({
                TableName: CACHE_DYNAMO_TABLE,
                IndexName: "revalidate",
                KeyConditionExpression: "#key = :key AND #revalidatedAt > :lastModified",
                ExpressionAttributeNames: {
                    "#key": "path",
                    "#revalidatedAt": "revalidatedAt",
                },
                ExpressionAttributeValues: {
                    ":key": { S: this.buildDynamoKey(key) },
                    ":lastModified": { N: String(lastModified ?? 0) },
                },
            }));
            const revalidatedTags = result.Items ?? [];
            debug("revalidatedTags", revalidatedTags);
            // If we have revalidated tags we return -1 to force revalidation
            return revalidatedTags.length > 0 ? -1 : lastModified ?? Date.now();
        }
        catch (e) {
            error("Failed to get revalidated tags", e);
            return lastModified ?? Date.now();
        }
    }
    async getByTag(tag) {
        try {
            if (disableDynamoDBCache)
                return [];
            const { Items } = await this.dynamoClient.send(new QueryCommand({
                TableName: CACHE_DYNAMO_TABLE,
                KeyConditionExpression: "#tag = :tag",
                ExpressionAttributeNames: {
                    "#tag": "tag",
                },
                ExpressionAttributeValues: {
                    ":tag": { S: this.buildDynamoKey(tag) },
                },
            }));
            return (
            // We need to remove the buildId from the path
            Items?.map(({ path: { S: key } }) => key?.replace(`${this.buildId}/`, "") ?? "") ?? []);
        }
        catch (e) {
            error("Failed to get by tag", e);
            return [];
        }
    }
    async batchWriteDynamoItem(req) {
        try {
            if (disableDynamoDBCache)
                return;
            await Promise.all(this.chunkArray(req, 25).map((Items) => {
                return this.dynamoClient.send(new BatchWriteItemCommand({
                    RequestItems: {
                        [CACHE_DYNAMO_TABLE ?? ""]: Items.map((Item) => ({
                            PutRequest: {
                                Item: {
                                    ...this.buildDynamoObject(Item.path, Item.tag),
                                },
                            },
                        })),
                    },
                }));
            }));
        }
        catch (e) {
            error("Failed to batch write dynamo item", e);
        }
    }
    buildDynamoKey(key) {
        // FIXME: We should probably use something else than path.join here
        // this could transform some fetch cache key into a valid path
        return path.posix.join(this.buildId, key);
    }
    buildDynamoObject(path, tags) {
        return {
            path: { S: this.buildDynamoKey(path) },
            tag: { S: this.buildDynamoKey(tags) },
            revalidatedAt: { N: `${Date.now()}` },
        };
    }
    chunkArray(array, chunkSize) {
        const chunks = [];
        for (let i = 0; i < array.length; i += chunkSize) {
            chunks.push(array.slice(i, i + chunkSize));
        }
        return chunks;
    }
    // S3 handling
    buildS3Key(key, extension) {
        return path.posix.join(CACHE_BUCKET_KEY_PREFIX ?? "", extension === "fetch" ? "__fetch" : "", this.buildId, extension === "fetch" ? key : `${key}.${extension}`);
    }
    buildS3KeyPrefix(key) {
        return path.posix.join(CACHE_BUCKET_KEY_PREFIX ?? "", this.buildId, key);
    }
    async listS3Object(key) {
        const { Contents } = await this.client.send(new ListObjectsV2Command({
            Bucket: CACHE_BUCKET_NAME,
            // add a point to the key so that it only matches the key and
            // not other keys starting with the same string
            Prefix: `${this.buildS3KeyPrefix(key)}.`,
        }));
        return (Contents ?? []).map(({ Key }) => Key);
    }
    async getS3Object(key, extension, keys) {
        try {
            if (!keys.includes(this.buildS3Key(key, extension)))
                return { Body: null, LastModified: null };
            const result = await this.client.send(new GetObjectCommand({
                Bucket: CACHE_BUCKET_NAME,
                Key: this.buildS3Key(key, extension),
            }));
            return result;
        }
        catch (e) {
            error("This error can usually be ignored : ", e);
            return { Body: null, LastModified: null };
        }
    }
    putS3Object(key, extension, value) {
        return this.client.send(new PutObjectCommand({
            Bucket: CACHE_BUCKET_NAME,
            Key: this.buildS3Key(key, extension),
            Body: value,
        }));
    }
    async deleteS3Objects(key) {
        try {
            const regex = new RegExp(`\.(json|rsc|html|body|meta|fetch|redirect)$`);
            const s3Keys = (await this.listS3Object(key)).filter((key) => key && regex.test(key));
            await this.client.send(new DeleteObjectsCommand({
                Bucket: CACHE_BUCKET_NAME,
                Delete: {
                    Objects: s3Keys.map((Key) => ({ Key })),
                },
            }));
        }
        catch (e) {
            error("Failed to delete cache", e);
        }
    }
}
