import { compile, match } from "path-to-regexp";
import { NextConfig } from "../config";
import { debug } from "../logger";
import { escapeRegex, unescapeRegex } from "../util";
import { convertQuery, getUrlParts, isExternal } from "./util";
const routeHasMatcher = (headers, cookies, query) => (redirect) => {
    switch (redirect.type) {
        case "header":
            return (!!headers?.[redirect.key.toLowerCase()] &&
                new RegExp(redirect.value ?? "").test(headers[redirect.key.toLowerCase()] ?? ""));
        case "cookie":
            return (!!cookies?.[redirect.key] &&
                new RegExp(redirect.value ?? "").test(cookies[redirect.key] ?? ""));
        case "query":
            return query[redirect.key] && Array.isArray(redirect.value)
                ? redirect.value.reduce((prev, current) => prev || new RegExp(current).test(query[redirect.key]), false)
                : new RegExp(redirect.value ?? "").test(query[redirect.key] ?? "");
        case "host":
            return (headers?.host !== "" &&
                new RegExp(redirect.value ?? "").test(headers.host));
        default:
            return false;
    }
};
function checkHas(matcher, has, inverted = false) {
    return has
        ? has.reduce((acc, cur) => {
            if (acc === false)
                return false;
            return inverted ? !matcher(cur) : matcher(cur);
        }, true)
        : true;
}
function convertMatch(match, toDestination, destination) {
    if (match) {
        const { params } = match;
        const isUsingParams = Object.keys(params).length > 0;
        if (isUsingParams) {
            return toDestination(params);
        }
        else {
            return destination;
        }
    }
    else {
        return destination;
    }
}
export function addNextConfigHeaders(event, configHeaders) {
    const addedHeaders = {};
    if (!configHeaders)
        return addedHeaders;
    const { rawPath, headers, query, cookies } = event;
    const matcher = routeHasMatcher(headers, cookies, query);
    const requestHeaders = {};
    for (const { headers, has, missing, regex, source } of configHeaders) {
        if (new RegExp(regex).test(rawPath) &&
            checkHas(matcher, has) &&
            checkHas(matcher, missing, true)) {
            const fromSource = match(source);
            const _match = fromSource(rawPath);
            headers.forEach((h) => {
                const key = convertMatch(_match, compile(h.key), h.key);
                const value = convertMatch(_match, compile(h.value), h.value);
                requestHeaders[key] = value;
            });
        }
    }
    return requestHeaders;
}
export function handleRewrites(event, rewrites) {
    const { rawPath, headers, query, cookies } = event;
    const matcher = routeHasMatcher(headers, cookies, query);
    const rewrite = rewrites.find((route) => new RegExp(route.regex).test(rawPath) &&
        checkHas(matcher, route.has) &&
        checkHas(matcher, route.missing, true));
    const urlQueryString = new URLSearchParams(convertQuery(query)).toString();
    let rewrittenUrl = rawPath;
    const isExternalRewrite = isExternal(rewrite?.destination);
    debug("isExternalRewrite", isExternalRewrite);
    if (rewrite) {
        const { pathname, protocol, hostname } = getUrlParts(rewrite.destination, isExternalRewrite);
        const toDestination = compile(escapeRegex(pathname ?? "") ?? "");
        const fromSource = match(escapeRegex(rewrite?.source) ?? "");
        const _match = fromSource(rawPath);
        if (_match) {
            const { params } = _match;
            const isUsingParams = Object.keys(params).length > 0;
            if (isUsingParams) {
                const rewrittenPath = unescapeRegex(toDestination(params));
                rewrittenUrl = isExternalRewrite
                    ? `${protocol}//${hostname}${rewrittenPath}`
                    : `${rewrittenPath}`;
            }
            else {
                rewrittenUrl = rewrite.destination;
            }
            debug("rewrittenUrl", rewrittenUrl);
        }
    }
    const queryString = urlQueryString ? `?${urlQueryString}` : "";
    return {
        internalEvent: {
            ...event,
            rawPath: rewrittenUrl,
            url: `${rewrittenUrl}${queryString}`,
        },
        __rewrite: rewrite,
        isExternalRewrite,
    };
}
export function handleRedirects(event, redirects) {
    if (NextConfig.trailingSlash &&
        !event.rawPath.endsWith("/") &&
        !event.headers["x-nextjs-data"]) {
        return {
            type: event.type,
            statusCode: 308,
            headers: {
                Location: event.url + "/",
            },
            body: "",
            isBase64Encoded: false,
        };
        // eslint-disable-next-line sonarjs/elseif-without-else
    }
    else if (!NextConfig.trailingSlash &&
        event.rawPath.endsWith("/") &&
        event.rawPath !== "/") {
        return {
            type: event.type,
            statusCode: 308,
            headers: {
                Location: event.url.replace(/\/$/, ""),
            },
            body: "",
            isBase64Encoded: false,
        };
    }
    const { internalEvent, __rewrite } = handleRewrites(event, redirects.filter((r) => !r.internal));
    if (__rewrite && !__rewrite.internal) {
        return {
            type: event.type,
            statusCode: __rewrite.statusCode ?? 308,
            headers: {
                Location: internalEvent.url,
            },
            body: "",
            isBase64Encoded: false,
        };
    }
}
export function fixDataPage(internalEvent, buildId) {
    const { rawPath, query } = internalEvent;
    const dataPattern = `/_next/data/${buildId}`;
    if (rawPath.startsWith(dataPattern) && rawPath.endsWith(".json")) {
        const newPath = rawPath.replace(dataPattern, "").replace(/\.json$/, "");
        query.__nextDataReq = "1";
        const urlQuery = {};
        Object.keys(query).forEach((k) => {
            const v = query[k];
            urlQuery[k] = Array.isArray(v) ? v.join(",") : v;
        });
        return {
            ...internalEvent,
            rawPath: newPath,
            query,
            url: `${newPath}${query ? `?${new URLSearchParams(urlQuery).toString()}` : ""}`,
        };
    }
    return internalEvent;
}
export function handleFallbackFalse(internalEvent, prerenderManifest) {
    const { rawPath } = internalEvent;
    const { dynamicRoutes, routes } = prerenderManifest;
    const routeFallback = Object.entries(dynamicRoutes)
        .filter(([, { fallback }]) => fallback === false)
        .some(([, { routeRegex }]) => {
        const routeRegexExp = new RegExp(routeRegex);
        return routeRegexExp.test(rawPath);
    });
    if (routeFallback && !Object.keys(routes).includes(rawPath)) {
        return {
            ...internalEvent,
            rawPath: "/404",
            url: "/404",
            headers: {
                ...internalEvent.headers,
                "x-invoke-status": "404",
            },
        };
    }
    return internalEvent;
}
